package com.americancoders.edi;


import java.util.*;
import java.io.*;
import java.net.*;

import com.americancoders.edi.*;
import com.americancoders.edi.x12.*;
import org.w3c.dom.*;

import org.xml.sax.*;

/**
 * class builds OBOE objects by parsing input string in valid xml edi format
 *
 *<P>OBOE - Open Business Objects for EDI
 *<br>An EDI and XML Translator Written In Java
 *<br>Copyright 1998-2000 - American Coders, LTD  - Raleigh NC USA
 *<br>All rights reserved
 *<br>American Coders, Ltd
 *<br>P. O. Box 97462
 *<br>Raleigh, NC  27624  USA
 *<br>1-919-846-2014
 *<br>http://www.americancoders.com
@author Joe McVerry, American Coders Ltd.
@version 1.5.0
 *
 */

public class ValidXMLEDIParser extends org.xml.sax.helpers.DefaultHandler

{
  protected TransactionSet parsedTransactionSet;
  protected int _iElement = 0;
  protected int _iLine = 0;
  protected Table table = null;
  protected Table headerTable = null;
  protected Table detailTable = null;
  protected Table summaryTable = null;
  protected Segment segment = null;
  protected CompositeDE composite = null;
  protected DataElement element = null;
  protected SegmentContainer segmentContainer = null;
  protected ICompositeDEContainer compositeContainer = null;
  protected IDataElementContainer elementContainer = null;
  protected Stack segmentStack = new Stack();
  protected int elementCount = 0;
  protected CharArrayWriter contents = new CharArrayWriter();
  protected org.apache.xerces.parsers.SAXParser parser;

  /**
   * create a transaction set from input string
   * @param inString the edi document
   * @exception Exception
   *                      - unknown transaction set, this transaction set is undefined to OBOE
   */

  public ValidXMLEDIParser()
  throws Exception
  {
      parser = new org.apache.xerces.parsers.SAXParser();

      try {
        parser.setFeature("http://xml.org/sax/features/validation", true);
        parser.setFeature("http://xml.org/sax/features/namespaces", false);
      }
      catch (SAXException e) {System.out.println("error in setting up parser feature");}

      parser.setContentHandler(this);
      parser.setErrorHandler(this);
  }

  /**
   * parse an xml document coming in as a String
   * @param String
   * @exception Exception
   *                      - as thrown by the parser
   */
  public void parse(String text)
  throws Exception
  {
    CharArrayReader car = new CharArrayReader(text.toCharArray());
    InputSource is = new InputSource(car);
    Properties properties = new Properties();
    FileInputStream inStream = new FileInputStream(TransactionSetFactory.getPropertiesFilePath());
    properties.load(inStream);
    is.setSystemId("file:///"+properties.getProperty("xmlPath"));
    parser.parse(is);
  }

  /**
   * parse an xml
   * @param String - incoming file name
   * @exception Exception
   *                      - as thrown by the parser
   */
  public void parseFile(String filename)
  throws Exception
  {
    InputSource is = new InputSource("file:///"+filename);
    parser.parse(is);
  }

  /**
   * method called for each xml element found.
   * <br> process logic
   * <ul>
   * <li> based on the name of the element found
   * <li> for each pull appropriate attributes and construct object
   * <li> if owned by another class, and all are except for Standard, add it to its parents object
   * </ul>
   * @param String data element name
   * @param AttributeList attributes associated with data element
   *
   */
  public void startElement(java.lang.String uri,
  java.lang.String localName,
  java.lang.String rawName,
  Attributes attributes)
  throws SAXException
  {
    int i;

    _iElement++;
    contents.reset();

    String name = rawName;
    // Standard

    try {
      if (name.compareTo("") == 0) {
        return;
      }
      else
      if (name.compareTo("transactionset") == 0) {
        String tsID = attributes.getValue("code");
        parsedTransactionSet = TransactionSetFactory.buildTransactionSet(tsID);
      }
      else
      if (name.compareTo("table") == 0) {
        String section = attributes.getValue("section");
        if (section.compareTo("header") == 0) {
          if (headerTable != null)
          throw new SAXException("header table already defined");
          headerTable = table = parsedTransactionSet.getHeaderTable();
          segmentContainer = table;
        }
        else
        if (section.compareTo("detail") == 0) {
          if (detailTable != null)
          throw new SAXException("detail table already defined");
          detailTable = table = parsedTransactionSet.getDetailTable();
          segmentContainer = table;
        }
        else
        if (section.compareTo("summary") == 0) {
          if (summaryTable != null)
          throw new SAXException("summary table already defined");
          summaryTable = table = parsedTransactionSet.getSummaryTable();
          segmentContainer = table;
        }
        else
        throw new SAXException("table section="+section+" unknown");
      }
      else
      if (name.compareTo("segment") == 0) {
        String code = attributes.getValue("code");
        TemplateSegment templateSegment = segmentContainer.getTemplateSegment(code);
        segment = new Segment(templateSegment);
        segmentContainer.addSegment(segment);
        segmentStack.push(segmentContainer);
        segmentStack.push(new Integer(elementCount));
        elementCount=0;
        segmentContainer = segment;
        compositeContainer = segment;
        elementContainer = segment;
      }
      else
      if (name.compareTo("composite") == 0) {
        String code = attributes.getValue("code");
        TemplateSegment templateSegment = segment.getTemplate();
        TemplateComposite templateComposite = templateSegment.getTemplateComposite(code);
        composite = new CompositeDE(templateComposite);
        compositeContainer.addCompositeDE(composite);
        elementCount++;
        segmentStack.push(segmentContainer);
        segmentStack.push(new Integer(elementCount));
        elementContainer = composite;
        elementCount=0;
      }
      else
      if (name.compareTo("element") == 0) {
        String code = attributes.getValue("code");
        if (elementContainer instanceof CompositeDE)
        {
          TemplateComposite templateComposite = composite.getTemplate();
          TemplateDE templateDE = templateComposite.getTemplateDE(elementCount);
          if (templateDE.getID().compareTo(code) != 0)
          throw new SAXException("element " + code + " not in correct position or composite");
          composite.buildDE(elementCount);
          element = composite.getDataElement(elementCount);
          elementCount++;
        }
        else
        {
          TemplateSegment templateSegment = segment.getTemplate();
          TemplateDE templateDE = templateSegment.getTemplateDE(elementCount);
          if (templateDE.getID().compareTo(code) != 0)
          throw new SAXException("element " + code + " not in correct position or segment");
          segment.buildDE(elementCount);
          element = segment.getDataElement(elementCount);
          elementCount++;
        }
        element.set("");

      }
      else
      if (name.compareTo("value") == 0) {

      }

      else System.err.println("Logic error: Unknown element name \"" + name
      + "\" found at element position " + _iElement + " line: " + _iLine);

    }
  catch (Exception e1) {e1.printStackTrace();}


  }

  /**
   * Method declaration
   *
   *
   * @param name
   *
   * @throws SAXException
   *
   *
   */
  public void endElement(java.lang.String uri,
  java.lang.String localName,
  java.lang.String rawName)
  throws SAXException{
    String name = rawName;
    Integer intObject;
    Object obj;
    if (name.compareTo("dataelement") == 0) {
      element.set(contents.toString());
    }
    else if (name.compareTo("value") == 0) {
      element.set(contents.toString());
    }
    else if (name.compareTo("composite") == 0) {
      intObject = (Integer) segmentStack.pop();
      elementCount = intObject.intValue();
      obj = segmentStack.pop();
      elementContainer = segment;

    }
    else if (name.compareTo("segment") == 0) {
      intObject = (Integer) segmentStack.pop();
      elementCount = intObject.intValue();
      obj = segmentStack.pop();
      if (obj instanceof Table)
      {
        segmentContainer = (SegmentContainer) obj;
      }
      else   {
        segment = (Segment) obj;
        segmentContainer = segment;
        compositeContainer = segment;
        elementContainer = segment;
      }
    }
  }

  /**
   * Method declaration
   *
   *
   * @param ch
   * @param start
   * @param length
   *
   *
   */
  public void characters(char ch[], int start, int length) {
    contents.write(ch, start, length);
  }

  public void ignorableWhitespace(char[] ch, int start, int length)
  {
    for (int i=start; i < start+length; i++)
    if (ch[i] == '\n') _iLine++;
  }

  /**
   * catches warning SAXParseExceptions
   * this code sends exception to stdio and allows public classto continue
   */
  public void warning(SAXParseException e) throws SAXException {
    System.err.println("Warning at (file " + e.getSystemId() + ", line "
    + e.getLineNumber() + ", char "
    + e.getColumnNumber() + "): " + e.getMessage());
  }

  /**
   * catches error SAXParseExceptions
   * this code causes exception to continue
   */
  public void error(SAXParseException e) throws SAXException {
    throw new SAXException("Error at (file " + e.getSystemId()
    + ", line " + e.getLineNumber() + ", char "
    + e.getColumnNumber() + "): "
    + e.getMessage());
  }

  /**
   * catches fatal SAXParseExceptions
   * this code causes exception to continue
   */
  public void fatalError(SAXParseException e) throws SAXException {
    throw new SAXException("Fatal Error at (file " + e.getSystemId()
    + ", line " + e.getLineNumber() + ", char "
    + e.getColumnNumber() + "): "
    + e.getMessage());
  }
  /**
   * static class method will build a transaction set based on input string and OBOE.properties defintion
   * see OBOE.properties file to define the directory path
   * @param String transaction id
   * @return TransactionSet
   */
  public static TransactionSet buildTransactionSet(String inTSID)
  throws Exception
  {

    return null;
  }


  /**
   * static main method to test xmlparser class
   *
   */
  public static void main(String args[])
  {
    try {

      ValidXMLEDIParser xmlp = new ValidXMLEDIParser();
      xmlp.parseFile(args[0]);

      System.out.println(xmlp.getTransactionSet().getFormattedText(TransactionSet.X12_FORMAT));
    }
    catch (Exception e1)
    {
      e1.printStackTrace();
    }
  }


  /**
   * returns the transaction set that was parsed
   * @return TransactionSet - the transaction set parsed when object was built.
   */
  public TransactionSet getTransactionSet()
  {
    return parsedTransactionSet;
  }

}
